package de.fhdw.wtf.common.ast.type;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;

import de.fhdw.wtf.common.ast.UnqualifiedName;
import de.fhdw.wtf.common.ast.visitor.AtomicTypeVisitorReturn;
import de.fhdw.wtf.common.ast.visitor.CompositeTypeVisitor;
import de.fhdw.wtf.common.ast.visitor.CompositeTypeVisitorException;
import de.fhdw.wtf.common.ast.visitor.CompositeTypeVisitorReturn;
import de.fhdw.wtf.common.ast.visitor.CompositeTypeVisitorReturnException;
import de.fhdw.wtf.common.ast.visitor.TypeProxyStateVisitorException;
import de.fhdw.wtf.common.ast.visitor.TypeVisitorException;
import de.fhdw.wtf.common.ast.visitor.TypeVisitorReturn;
import de.fhdw.wtf.common.constants.parser.AstDescriptionConstants;
import de.fhdw.wtf.common.constants.referencer.BaseTypeConstants;
import de.fhdw.wtf.common.exception.ast.ASTException;
import de.fhdw.wtf.common.exception.parser.SumIsAnythingException;
import de.fhdw.wtf.common.token.DummyToken;
import de.fhdw.wtf.common.token.IdentifierToken;
import de.fhdw.wtf.common.token.Token;
import de.fhdw.wtf.walker.walker.HelperUtils;

/**
 * A {@link SumType} is a synthetic SuperType for all {@link Type}s in the Sum.<br/>
 * For example: An <Operation> with a Sum of two Classes as returnValue (The two classes are not in the same inheritance
 * structure). Normally you have to make a abstract class or interface, which both classes extends or implements.
 */
public class SumType extends CompositeType {
	
	/**
	 * generated.
	 */
	private static final long serialVersionUID = -7593357358171816032L;
	
	/**
	 * The Elements.
	 */
	private final List<Type> elements;
	
	/**
	 * Constructor of {@link SumType}.
	 * 
	 * @param firstToken
	 *            firstToken
	 * @param lastToken
	 *            lastToken
	 */
	protected SumType(final Token firstToken, final Token lastToken) {
		super(firstToken, lastToken);
		this.elements = new ArrayList<>();
	}
	
	/**
	 * Creates a {@link SumType}-Object.
	 * 
	 * @param firstToken
	 *            firstToken
	 * @param lastToken
	 *            lastToken
	 * @return the {@link SumType}-Object
	 */
	public static SumType create(final Token firstToken, final Token lastToken) {
		return new SumType(firstToken, lastToken);
	}
	
	/**
	 * Creates a {@link SumType}-Object.
	 * 
	 * @param firstToken
	 *            firstToken
	 * @return the {@link SumType}-Object
	 */
	public static SumType create(final Token firstToken) {
		return new SumType(firstToken, null);
	}
	
	/**
	 * Adds a type to the list of elements.
	 * 
	 * @param to
	 *            type
	 */
	public void add(final Type to) {
		this.elements.add(to);
	}
	
	/**
	 * Returns the list of elements.
	 * 
	 * @return elements
	 */
	public List<Type> getElements() {
		return this.elements;
	}
	
	@Override
	public int compareTo(final Type o) {
		return o.accept(new TypeVisitorReturn<Integer>() {
			
			@Override
			public Integer handle(final AtomicType s) {
				return s.accept(new AtomicTypeVisitorReturn<Integer>() {
					
					@Override
					public Integer handle(final BaseType baseType) {
						return 1;
					}
					
					@Override
					public Integer handle(final ClassType clazz) {
						return -1;
					}
				});
			}
			
			@Override
			public Integer handle(final CompositeType c) {
				return c.accept(new CompositeTypeVisitorReturn<Integer>() {
					
					@Override
					public Integer handle(final ListType list) {
						return 1;
					}
					
					@Override
					public Integer handle(final MapType map) {
						return 1;
					}
					
					@Override
					public Integer handle(final ProductType product) {
						return -1;
					}
					
					@Override
					public Integer handle(final SumType sum) {
						if (SumType.this.getElements().size() > sum.getElements().size()) {
							return 1;
						}
						if (SumType.this.getElements().size() < sum.getElements().size()) {
							return -1;
						}
						
						final List<Type> thisSorted = SumType.this.sortElements();
						final List<Type> otherSorted = sum.sortElements();
						
						for (int i = 0; i < thisSorted.size(); i++) {
							final Type thisElement = thisSorted.get(i);
							final Type otherElement = otherSorted.get(i);
							final int currentCompare = thisElement.compareTo(otherElement);
							if (currentCompare != 0) {
								return currentCompare;
							}
						}
						
						return 0;
					}
					
					@Override
					public Integer handle(final ThrownType thrownType) {
						return 1;
					}
				});
			}
			
			@Override
			public Integer handle(final TypeProxy s) {
				return 1;
			}
		});
	}
	
	/**
	 * This method returns a sorted list of the elements of <code>this</code>.<br>
	 * 
	 * @return ArrayList of Type.
	 */
	public List<Type> sortElements() {
		final List<Type> copy = new ArrayList<>();
		copy.addAll(this.getElements());
		Collections.sort(copy);
		return copy;
	}
	
	@Override
	public boolean equals(final Object o) {
		if (o instanceof SumType) {
			final SumType other = (SumType) o;
			if (this.getElements().size() != other.getElements().size()) {
				return false;
			}
			final List<Type> thisElements = this.getElements();
			final List<Type> otherElements = other.getElements();
			for (int i = 0; i < thisElements.size(); i++) {
				if (!thisElements.get(i).equals(otherElements.get(i))) {
					return false;
				}
			}
			return true;
		}
		return false;
	}
	
	@Override
	public int hashCode() {
		int result = 0;
		for (final Type type : this.getElements()) {
			result ^= type.hashCode();
		}
		return result;
	}
	
	@Override
	public String toString() {
		final StringBuilder result = new StringBuilder();
		result.append(AstDescriptionConstants.CURLY_BRACKET_OPEN_TOKEN);
		final Iterator<Type> it = this.elements.iterator();
		while (it.hasNext()) {
			result.append(it.next());
			if (it.hasNext()) {
				result.append(',');
			}
		}
		result.append(AstDescriptionConstants.CURLY_BRACKET_CLOSE_TOKEN);
		return result.toString();
	}
	
	@Override
	public String getTypeString() {
		final StringBuilder result = new StringBuilder();
		result.append(AstDescriptionConstants.CURLY_BRACKET_OPEN_TOKEN);
		final Iterator<Type> it = this.elements.iterator();
		while (it.hasNext()) {
			result.append(it.next().getTypeString());
			if (it.hasNext()) {
				result.append(',');
			}
		}
		result.append(AstDescriptionConstants.CURLY_BRACKET_CLOSE_TOKEN);
		return result.toString();
	}
	
	@Override
	public void accept(final CompositeTypeVisitor v) {
		v.handle(this);
	}
	
	@Override
	public <X> X accept(final CompositeTypeVisitorReturn<X> v) {
		return v.handle(this);
	}
	
	@Override
	public <X extends Exception> void accept(final CompositeTypeVisitorException<X> v) throws X {
		v.handle(this);
	}
	
	@Override
	public <X, Y extends Exception> X accept(final CompositeTypeVisitorReturnException<X, Y> v) throws Y {
		return v.handle(this);
	}
	
	@Override
	public boolean isSupertypeOf(final Type type) {
		if (this.equals(type)) {
			return false;
		}
		
		final List<Type> normSumElements;
		try {
			
			normSumElements = this.standardize().accept(new TypeVisitorReturn<List<Type>>() {
				
				@Override
				public List<Type> handle(final AtomicType atomicType) {
					throw new ASTException("A atomicType cannot be a standardized type of a sumType.");
				}
				
				@Override
				public List<Type> handle(final CompositeType compositeType) {
					return compositeType.accept(new CompositeTypeVisitorReturn<List<Type>>() {
						
						@Override
						public List<Type> handle(final ListType list) {
							throw new ASTException("A listType cannot be a standardized type of a sumType.");
						}
						
						@Override
						public List<Type> handle(final MapType map) {
							throw new ASTException("A mapType cannot be a standardized type of a sumType.");
						}
						
						@Override
						public List<Type> handle(final ProductType product) {
							throw new ASTException("A productType cannot be a standardized type of a sumType.");
						}
						
						@Override
						public List<Type> handle(final SumType sum) {
							return sum.getElements();
						}
						
						@Override
						public List<Type> handle(final ThrownType thrownType) {
							throw new ASTException("A thrownType cannot be a standardized type of a sumType.");
						}
					});
				}
				
				@Override
				public List<Type> handle(final TypeProxy typeProxy) {
					throw new ASTException("A typeType cannot be a standardized type of a sumType.");
				}
			});
		} catch (final SumIsAnythingException e) {
			return true;
		}
		
		return type.accept(new TypeVisitorReturn<Boolean>() {
			
			@Override
			public Boolean handle(final AtomicType atomicType) {
				for (final Type typeThis : normSumElements) {
					if (atomicType.equals(typeThis)) {
						return true;
					} else if (typeThis.isSupertypeOf(atomicType)) {
						return true;
					}
				}
				return false;
			}
			
			@Override
			public Boolean handle(final CompositeType compositeType) {
				return compositeType.accept(new CompositeTypeVisitorReturn<Boolean>() {
					
					@Override
					public Boolean handle(final ListType list) {
						return normSumElements.contains(list);
					}
					
					@Override
					public Boolean handle(final MapType map) {
						return normSumElements.contains(map);
					}
					
					@Override
					public Boolean handle(final ProductType product) {
						return normSumElements.contains(product);
					}
					
					@Override
					public Boolean handle(final SumType sum) {
						final List<Type> normSumElementsOther;
						try {
							normSumElementsOther = sum.standardize().accept(new TypeVisitorReturn<List<Type>>() {
								
								@Override
								public List<Type> handle(final AtomicType atomicType) {
									throw new ASTException("A atomicType cannot be a standardized type of a sumType.");
								}
								
								@Override
								public List<Type> handle(final CompositeType compositeType2) {
									return compositeType2.accept(new CompositeTypeVisitorReturn<List<Type>>() {
										
										@Override
										public List<Type> handle(final ListType list) {
											throw new ASTException(
													"A listType cannot be a standardized type of a sumType.");
										}
										
										@Override
										public List<Type> handle(final MapType map) {
											throw new ASTException(
													"A mapType cannot be a standardized type of a sumType.");
										}
										
										@Override
										public List<Type> handle(final ProductType product) {
											throw new ASTException(
													"A productType cannot be a standardized type of a sumType.");
										}
										
										@Override
										public List<Type> handle(final SumType sum2) {
											return sum2.getElements();
										}
										
										@Override
										public List<Type> handle(final ThrownType thrownType) {
											throw new ASTException(
													"A thrownType cannot be a standardized type of a sumType.");
										}
									});
								}
								
								@Override
								public List<Type> handle(final TypeProxy typeProxy) {
									throw new ASTException("A typeType cannot be a standardized type of a sumType.");
								}
							});
						} catch (final SumIsAnythingException e) {
							return false;
						}
						
						for (final Type typeOther : normSumElementsOther) {
							boolean contains = false;
							for (final Type typeThis : normSumElements) {
								if (typeOther.equals(typeThis)) {
									contains = true;
								} else if (typeThis.isSupertypeOf(typeOther)) {
									contains = true;
								}
							}
							if (!contains) {
								return false;
							}
						}
						return true;
					}
					
					@Override
					public Boolean handle(final ThrownType thrownType) {
						return normSumElements.contains(thrownType);
					}
				});
			}
			
			@Override
			public Boolean handle(final TypeProxy typeProxy) {
				return SumType.this.isSupertypeOf(HelperUtils.getTargetType(typeProxy));
			}
		});
	}
	
	/**
	 * Returns the number of elements in this sum, which are no thrownTypes.
	 * 
	 * @return number of elements
	 */
	public int getElementsSizeWithoutThrownTypes() {
		
		int result = 0;
		
		final Iterator<Type> subElementsIterator = this.getElements().iterator();
		
		while (subElementsIterator.hasNext()) {
			final Type subElement = subElementsIterator.next();
			
			if (subElement.accept(new TypeVisitorReturn<Boolean>() {
				
				@Override
				public Boolean handle(final AtomicType atomicType) {
					return true;
				}
				
				@Override
				public Boolean handle(final CompositeType compositeType) {
					return compositeType.accept(new CompositeTypeVisitorReturn<Boolean>() {
						
						@Override
						public Boolean handle(final ListType list) {
							return true;
						}
						
						@Override
						public Boolean handle(final MapType map) {
							return true;
						}
						
						@Override
						public Boolean handle(final ProductType product) {
							return true;
						}
						
						@Override
						public Boolean handle(final SumType sum) {
							return true;
						}
						
						@Override
						public Boolean handle(final ThrownType thrownType) {
							return false;
						}
					});
				}
				
				@Override
				public Boolean handle(final TypeProxy typeProxy) {
					return true;
				}
			})) {
				result = result + 1;
			}
		}
		
		return result;
	}
	
	@Override
	public Type standardize() throws SumIsAnythingException {
		final List<Type> copy = new ArrayList<>();
		copy.addAll(this.getElements());
		
		// doppelte Elemente raus:
		final List<Type> withoutDoubleElements = new ArrayList<>();
		for (final Type element : copy) {
			if (!withoutDoubleElements.contains(element)) {
				withoutDoubleElements.add(element);
			}
		}
		
		// Summen: Assoziativität
		final List<Type> internSumElements = new ArrayList<>();
		for (final Iterator<Type> iterator = withoutDoubleElements.iterator(); iterator.hasNext();) {
			final Type type = iterator.next();
			type.accept(new TypeVisitorException<SumIsAnythingException>() {
				
				@Override
				public void handle(final AtomicType s) throws SumIsAnythingException {
					// Nothing
				}
				
				@Override
				public void handle(final CompositeType c) throws SumIsAnythingException {
					c.accept(new CompositeTypeVisitorException<SumIsAnythingException>() {
						
						@Override
						public void handle(final MapType map) throws SumIsAnythingException {
							// Nothing
						}
						
						@Override
						public void handle(final ListType list) throws SumIsAnythingException {
							// Nothing
						}
						
						@Override
						public void handle(final ProductType product) throws SumIsAnythingException {
							// Nothing
						}
						
						@Override
						public void handle(final SumType sum) throws SumIsAnythingException {
							iterator.remove();
							
							internSumElements.addAll(sum.standardize().accept(new TypeVisitorReturn<List<Type>>() {
								
								@Override
								public List<Type> handle(final AtomicType atomicType) {
									throw new ASTException("A atomicType cannot be a standardized type of a sumType.");
								}
								
								@Override
								public List<Type> handle(final CompositeType compositeType) {
									return compositeType.accept(new CompositeTypeVisitorReturn<List<Type>>() {
										
										@Override
										public List<Type> handle(final ListType list) {
											throw new ASTException(
													"A listType cannot be a standardized type of a sumType.");
										}
										
										@Override
										public List<Type> handle(final MapType map) {
											throw new ASTException(
													"A mapType cannot be a standardized type of a sumType.");
										}
										
										@Override
										public List<Type> handle(final ProductType product) {
											throw new ASTException(
													"A productType cannot be a standardized type of a sumType.");
										}
										
										@Override
										public List<Type> handle(final SumType sum2) {
											return sum2.getElements();
										}
										
										@Override
										public List<Type> handle(final ThrownType thrownType) {
											throw new ASTException(
													"A thrownType cannot be a standardized type of a sumType.");
										}
									});
								}
								
								@Override
								public List<Type> handle(final TypeProxy typeProxy) {
									throw new ASTException("A typeProxy cannot be a standardized type of a sumType.");
								}
							}));
						}
						
						@Override
						public void handle(final ThrownType thrownType) throws SumIsAnythingException {
							// nothing to do
						}
					});
				}
				
				@Override
				public void handle(final TypeProxy s) throws SumIsAnythingException {
					// Nothing
				}
			});
		}
		for (final Type element : internSumElements) {
			if (!withoutDoubleElements.contains(element)) {
				withoutDoubleElements.add(element);
			}
		}
		
		// If it contains Anything throw specified Exception
		for (final Type type : withoutDoubleElements) {
			type.accept(new TypeVisitorException<SumIsAnythingException>() {
				
				@Override
				public void handle(final AtomicType s) throws SumIsAnythingException {
					return;
				}
				
				@Override
				public void handle(final CompositeType c) throws SumIsAnythingException {
					return;
				}
				
				@Override
				public void handle(final TypeProxy s) throws SumIsAnythingException {
					s.getState().accept(new TypeProxyStateVisitorException<SumIsAnythingException>() {
						
						@Override
						public void handle(final ByNameState byName) throws SumIsAnythingException {
							if (byName.getName().equals(
									UnqualifiedName.create(IdentifierToken.create(
											BaseTypeConstants.ANYTHING_CONSTANT,
											DummyToken.getDummyPosition())))) {
								throw new SumIsAnythingException();
							}
						}
						
						@Override
						public void handle(final InvalidState invalid) throws SumIsAnythingException {
							return;
						}
						
						@Override
						public void handle(final ByReferenceState byType) throws SumIsAnythingException {
							if (byType.getTargetName().equals(
									UnqualifiedName.create(IdentifierToken.create(
											BaseTypeConstants.ANYTHING_CONSTANT,
											DummyToken.getDummyPosition())))) {
								throw new SumIsAnythingException();
							}
						}
					});
				}
			});
		}
		
		// Spezieller | Genereller
		final List<Type> standardizeElements = new ArrayList<>();
		for (final Type current : withoutDoubleElements) {
			boolean putItIn = true;
			for (final Iterator<Type> iterator2 = standardizeElements.iterator(); iterator2.hasNext();) {
				final Type standardizedElement = iterator2.next();
				if (standardizedElement.isSupertypeOf(current)) {
					// Wenn current spezieller als element, dann nicht
					// hineinpacken
					putItIn = false;
				}
				if (current.isSupertypeOf(standardizedElement)) {
					// Wenn current genereller als element, dann element
					// herausnehmen
					iterator2.remove();
				}
			}
			if (putItIn) {
				standardizeElements.add(current);
			}
		}
		
		Collections.sort(standardizeElements);
		
		final SumType result = new SumType(DummyToken.getInstance(), DummyToken.getInstance());
		result.getElements().addAll(standardizeElements);
		
		return result;
	}
	
	@Override
	public boolean isThrownType() {
		return false;
	}
}
