package de.fhdw.wtf.common.stream;

import java.util.ArrayList;
import java.util.List;

import de.fhdw.wtf.common.token.Token;

/**
 * In this class the tokens {@link de.fhdw.wtf.common.token.WhitespaceToken} and
 * {@link de.fhdw.wtf.common.token.CommentTextToken} will be add to the list but skipped in all output operations.
 */
public final class FilteredTokenStream implements TokenStream {
	
	/**
	 * the data as vector with tokens.
	 */
	private final List<Token> data;
	
	/**
	 * Constructor of {@link FilteredTokenStream}.
	 */
	private FilteredTokenStream() {
		this.data = new ArrayList<>();
	}
	
	/**
	 * Constructor of {@link FilteredTokenStream}.
	 * 
	 * @param data
	 *            the data as vector with tokens.
	 */
	private FilteredTokenStream(final List<Token> data) {
		this.data = data;
	}
	
	/**
	 * Creates a {@link FilteredTokenStream}-Object.
	 * 
	 * @return the {@link FilteredTokenStream}-Object
	 */
	public static FilteredTokenStream create() {
		return new FilteredTokenStream();
	}
	
	@Override
	public void add(final Token t) {
		this.data.add(t);
	}
	
	@Override
	public Token next() {
		Token result = this.data.get(0);
		this.data.remove(0);
		while (result.isCommentToken() || result.isWhitespaceToken()) {
			result = this.data.get(0);
			this.data.remove(0);
		}
		return result;
	}
	
	@Override
	public boolean hasNext() {
		for (final Token t : this.data) {
			if (!(t.isCommentToken() || t.isWhitespaceToken())) {
				return true;
			}
		}
		return false;
	}
	
	@Override
	public boolean removeFirst() {
		if (this.hasNext()) {
			Token result = this.data.get(0);
			this.data.remove(0);
			while (result.isWhitespaceToken() || result.isCommentToken()) {
				result = this.data.get(0);
				this.data.remove(0);
			}
			return true;
		}
		return false;
	}
	
	@Override
	public Token peek() {
		Token result = this.data.get(0);
		for (int i = 1; i < this.data.size(); i++) {
			if (!result.isCommentToken() && !result.isWhitespaceToken()) {
				return result;
			}
			result = this.data.get(i);
		}
		return result;
	}
	
	@Override
	public TokenStream copy() {
		final List<Token> copiedData = new ArrayList<>();
		copiedData.addAll(this.data);
		return new FilteredTokenStream(copiedData);
	}
}
