package de.fhdw.wtf.dsl.scanner.common;

import de.fhdw.wtf.common.token.Position;

/**
 * Represents a transition in the {@link Scanner} "Automata". The transition is defined by given "from"
 * {@link StateMatcher}, {@link InputMatcher} and some rules for the target {@link ScannerState} witch will be
 * implemented by subclasses.
 * 
 * <h2>Example:</h2> Any state that matches the "from" {@link StateMatcher} := f<br>
 * Any input that matches the {@link InputMatcher} := i<br>
 * target state := t<br>
 * <br>
 * <b>f - i -> t</b>
 */
public abstract class Delta {
	
	private final StateMatcher fromStateMatcher;
	private final InputMatcher inputMatcher;
	
	/**
	 * Constructor for {@link Delta}s.
	 * 
	 * @param fromStateMatcher
	 *            : original state
	 * @param inputMatcher
	 */
	protected Delta(final StateMatcher fromStateMatcher, final InputMatcher inputMatcher) {
		this.fromStateMatcher = fromStateMatcher;
		this.inputMatcher = inputMatcher;
	}
	
	private StateMatcher getFromStateMatcher() {
		return this.fromStateMatcher;
	}
	
	private InputMatcher getInputMatcher() {
		return this.inputMatcher;
	}
	
	/**
	 * Tests if the given combination of current {@link ScannerState} and input {@link Character} matches this
	 * {@link Delta}.
	 * 
	 * @param currentState
	 *            the current State of the {@link Scanner}
	 * @param input
	 *            the input to be tested
	 * @return the matching result
	 */
	public boolean match(final ScannerState currentState, final Character input) {
		return this.getFromStateMatcher().match(currentState) && this.getInputMatcher().match(input);
	}
	
	/**
	 * Returns the next {@link ScannerState} by using this {@link Delta}.
	 * 
	 * @param position
	 *            start position of the new {@link ScannerState}
	 * @param currentState
	 *            the current {@link ScannerState}
	 * @return the next {@link ScannerState}
	 */
	public abstract ScannerState getNextScannerState(Position position, ScannerState currentState);
}
