package de.fhdw.wtf.dsl.scanner.modelScanner;

import java.util.ArrayList;
import java.util.List;

import de.fhdw.wtf.common.token.Position;
import de.fhdw.wtf.dsl.scanner.common.CreatingDelta;
import de.fhdw.wtf.dsl.scanner.common.Delta;
import de.fhdw.wtf.dsl.scanner.common.InputMatcher;
import de.fhdw.wtf.dsl.scanner.common.LoopDelta;
import de.fhdw.wtf.dsl.scanner.common.Scanner;
import de.fhdw.wtf.dsl.scanner.common.ScannerState;
import de.fhdw.wtf.dsl.scanner.common.StateCreater;
import de.fhdw.wtf.dsl.scanner.common.StateMatcher;
import de.fhdw.wtf.dsl.scanner.modelScanner.state.CommentState;
import de.fhdw.wtf.dsl.scanner.modelScanner.state.InvalidState;
import de.fhdw.wtf.dsl.scanner.modelScanner.state.KeywordIdentifierState;
import de.fhdw.wtf.dsl.scanner.modelScanner.state.ScannerStateVisitor;
import de.fhdw.wtf.dsl.scanner.modelScanner.state.SymbolState;
import de.fhdw.wtf.dsl.scanner.modelScanner.state.WhitespaceState;

/**
 * {@link Scanner} implementation for the model language of WTF.
 */
public final class ModelDslScanner extends Scanner {
	
	/**
	 * Default number of whitespaces, who representing a tabulator.
	 */
	private static final int DEFAULT_TAB_WIDTH = 2;
	
	/**
	 * List of allowed Transitions between ScannerStates.
	 */
	private final List<Delta> deltas;
	
	/**
	 * Creates a new ModelDslScanner-Object with the given tabulator width.
	 * 
	 * @param tabulatorWidth
	 *            the number of whitespaces representing a tabulator
	 * 
	 * @return a new instance of ModelDslScanner
	 */
	public static ModelDslScanner create(final int tabulatorWidth) {
		return new ModelDslScanner(tabulatorWidth);
	}
	
	/**
	 * Creates a new ModelDslScanner-Object, where the tabulator width is default.
	 * 
	 * @return a new instance of ModelDslScanner
	 */
	public static ModelDslScanner create() {
		return new ModelDslScanner(DEFAULT_TAB_WIDTH);
	}
	
	/**
	 * Creates a new ModelDslScanner-Object with the given tabulator width.
	 * 
	 * @param tabulatorWidth
	 *            the number of whitespaces representing a tabulator
	 */
	private ModelDslScanner(final int tabulatorWidth) {
		super(tabulatorWidth);
		this.deltas = new ArrayList<>();
		this.initDelta();
	}
	
	@Override
	protected List<Delta> getDeltas() {
		return this.deltas;
	}
	
	@Override
	protected ScannerState getStartState() {
		return SymbolState.create(Position.create("", 1, 1, 0));
	}
	
	@Override
	protected ScannerState getDefaultState(final Position position) {
		return InvalidState.create(position);
	}
	
	/**
	 * Builds the delta-function for the 'Automata' the scanner uses.
	 */
	private void initDelta() {
		final StateCreater commentCreater = new StateCreater() {
			@Override
			public ScannerState create(final Position position) {
				return CommentState.create(position);
			}
		};
		final StateCreater whitespaceCreater = new StateCreater() {
			@Override
			public ScannerState create(final Position position) {
				return WhitespaceState.create(position);
			}
		};
		final StateCreater keyIdentCreater = new StateCreater() {
			@Override
			public ScannerState create(final Position position) {
				return KeywordIdentifierState.create(position);
			}
		};
		final StateCreater symbolCreater = new StateCreater() {
			@Override
			public ScannerState create(final Position position) {
				return SymbolState.create(position);
			}
		};
		final InputMatcher keyIdentPrefixMatcher = new InputMatcher() {
			@Override
			public boolean match(final Character c) {
				return Character.isLetter(c) || c == '_';
			}
		};
		final InputMatcher keyIdentMatcher = new InputMatcher() {
			@Override
			public boolean match(final Character c) {
				return Character.isLetterOrDigit(c) || c == '_';
			}
		};
		final InputMatcher whitespaceMatcher = new InputMatcher() {
			@Override
			public boolean match(final Character c) {
				return Character.isWhitespace(c);
			}
		};
		
		final InputMatcher invalidMatcher = new InputMatcher() {
			@Override
			public boolean match(final Character c) {
				return InvalidState.isInValid(c);
			}
		};
		final InputMatcher symbolMatcher = new InputMatcher() {
			@Override
			public boolean match(final Character c) {
				return SymbolState.isSymbol(c);
			}
		};
		
		final InputMatcher commentBeginMatcher = new InputMatcher() {
			@Override
			public boolean match(final Character c) {
				return c == CommentState.COMMENT_OPEN;
			}
		};
		final InputMatcher commentMatcher = new InputMatcher() {
			@Override
			public boolean match(final Character c) {
				return c != CommentState.COMMENT_CLOSE;
			}
		};
		final InputMatcher commentEndMatcher = new InputMatcher() {
			@Override
			public boolean match(final Character c) {
				return c == CommentState.COMMENT_CLOSE;
			}
		};
		final StateMatcher exceptCommentState = new StateMatcher() {
			@Override
			public boolean match(final ScannerState state) {
				return state.accept(new ScannerStateVisitor() {
					@Override
					public boolean handleWhitespaceState() {
						return true;
					}
					
					@Override
					public boolean handleSymbolState() {
						return true;
					}
					
					@Override
					public boolean handleKeywordIdentifierState() {
						return true;
					}
					
					@Override
					public boolean handleInvalidState() {
						return true;
					}
					
					@Override
					public boolean handleCommentState() {
						return false;
					}
				});
			}
		};
		final StateMatcher exceptCommentAndSymbolState = new StateMatcher() {
			@Override
			public boolean match(final ScannerState state) {
				return state.accept(new ScannerStateVisitor() {
					@Override
					public boolean handleWhitespaceState() {
						return true;
					}
					
					@Override
					public boolean handleSymbolState() {
						return false;
					}
					
					@Override
					public boolean handleKeywordIdentifierState() {
						return true;
					}
					
					@Override
					public boolean handleInvalidState() {
						return true;
					}
					
					@Override
					public boolean handleCommentState() {
						return false;
					}
				});
			}
		};
		final StateMatcher exceptCommentAndKeyIdentState = new StateMatcher() {
			@Override
			public boolean match(final ScannerState state) {
				return state.accept(new ScannerStateVisitor() {
					@Override
					public boolean handleWhitespaceState() {
						return true;
					}
					
					@Override
					public boolean handleSymbolState() {
						return true;
					}
					
					@Override
					public boolean handleKeywordIdentifierState() {
						return false;
					}
					
					@Override
					public boolean handleInvalidState() {
						return true;
					}
					
					@Override
					public boolean handleCommentState() {
						return false;
					}
				});
			}
		};
		final StateMatcher exceptCommentAndWhitespaceState = new StateMatcher() {
			@Override
			public boolean match(final ScannerState state) {
				return state.accept(new ScannerStateVisitor() {
					@Override
					public boolean handleWhitespaceState() {
						return false;
					}
					
					@Override
					public boolean handleSymbolState() {
						return true;
					}
					
					@Override
					public boolean handleKeywordIdentifierState() {
						return true;
					}
					
					@Override
					public boolean handleInvalidState() {
						return true;
					}
					
					@Override
					public boolean handleCommentState() {
						return false;
					}
				});
			}
		};
		final StateMatcher whitespaceState = new StateMatcher() {
			@Override
			public boolean match(final ScannerState state) {
				return state.accept(new ScannerStateVisitor() {
					@Override
					public boolean handleWhitespaceState() {
						return true;
					}
					
					@Override
					public boolean handleSymbolState() {
						return false;
					}
					
					@Override
					public boolean handleKeywordIdentifierState() {
						return false;
					}
					
					@Override
					public boolean handleInvalidState() {
						return false;
					}
					
					@Override
					public boolean handleCommentState() {
						return false;
					}
				});
			}
		};
		final StateMatcher invalidState = new StateMatcher() {
			@Override
			public boolean match(final ScannerState state) {
				return state.accept(new ScannerStateVisitor() {
					@Override
					public boolean handleWhitespaceState() {
						return false;
					}
					
					@Override
					public boolean handleSymbolState() {
						return false;
					}
					
					@Override
					public boolean handleKeywordIdentifierState() {
						return false;
					}
					
					@Override
					public boolean handleInvalidState() {
						return true;
					}
					
					@Override
					public boolean handleCommentState() {
						return false;
					}
				});
			}
		};
		final StateMatcher keyIdentState = new StateMatcher() {
			@Override
			public boolean match(final ScannerState state) {
				return state.accept(new ScannerStateVisitor() {
					@Override
					public boolean handleWhitespaceState() {
						return false;
					}
					
					@Override
					public boolean handleSymbolState() {
						return false;
					}
					
					@Override
					public boolean handleKeywordIdentifierState() {
						return true;
					}
					
					@Override
					public boolean handleInvalidState() {
						return false;
					}
					
					@Override
					public boolean handleCommentState() {
						return false;
					}
				});
			}
		};
		final StateMatcher commentState = new StateMatcher() {
			@Override
			public boolean match(final ScannerState state) {
				return state.accept(new ScannerStateVisitor() {
					@Override
					public boolean handleWhitespaceState() {
						return false;
					}
					
					@Override
					public boolean handleSymbolState() {
						return false;
					}
					
					@Override
					public boolean handleKeywordIdentifierState() {
						return false;
					}
					
					@Override
					public boolean handleInvalidState() {
						return false;
					}
					
					@Override
					public boolean handleCommentState() {
						return true;
					}
				});
			}
		};
		final StateMatcher symbolState = new StateMatcher() {
			@Override
			public boolean match(final ScannerState state) {
				return state.accept(new ScannerStateVisitor() {
					@Override
					public boolean handleWhitespaceState() {
						return false;
					}
					
					@Override
					public boolean handleSymbolState() {
						return true;
					}
					
					@Override
					public boolean handleKeywordIdentifierState() {
						return false;
					}
					
					@Override
					public boolean handleInvalidState() {
						return false;
					}
					
					@Override
					public boolean handleCommentState() {
						return false;
					}
				});
			}
		};
		this.getDeltas().add(CreatingDelta.create(exceptCommentAndSymbolState, symbolMatcher, symbolCreater));
		this.getDeltas().add(
				CreatingDelta.create(exceptCommentAndKeyIdentState, keyIdentPrefixMatcher, keyIdentCreater));
		this.getDeltas().add(
				CreatingDelta.create(exceptCommentAndWhitespaceState, whitespaceMatcher, whitespaceCreater));
		this.getDeltas().add(LoopDelta.create(invalidState, invalidMatcher));
		this.getDeltas().add(LoopDelta.create(whitespaceState, whitespaceMatcher));
		this.getDeltas().add(LoopDelta.create(keyIdentState, keyIdentMatcher));
		this.getDeltas().add(LoopDelta.create(symbolState, symbolMatcher));
		this.getDeltas().add(CreatingDelta.create(exceptCommentState, commentBeginMatcher, commentCreater));
		this.getDeltas().add(CreatingDelta.create(commentState, commentEndMatcher, symbolCreater));
		this.getDeltas().add(LoopDelta.create(commentState, commentMatcher));
	}
	
}
